--[[
Restoration Shaman Stat Analyzer, by Rheeah

Expanded and enhanced version based on the original "Resto Shaman Mastery Analyzer" addon by _swift
See:
http://www.curse.com/addons/wow/rshamanmastery

--]]

local revision = tonumber(string.sub("$Revision: 12 $", 12, -3))

local effectiveTotal = 0.0               -- total effective healing in current combat
local weightedTotal = 0.0                -- total weighted by the mastery effect of each heal
local masteryTotal = 0.0                 -- effective healing from Mastery in current combat
local criticalTotal = 0.0                -- effective healing from Critical in current combat
local incrementalMastery = 0.0           -- increase to effective healing from +1 Mastery Rating
local incrementalCritical = 0.0          -- increase to effective healing from +1 Critical Rating
local lastCast = { }                     -- table of most recent cast of each spell (indexed by spell name), for matching with AA
local healthTable = { }                  -- table of most recent health and max health of every known friendly unit
local spiritLinkGUID = nil               -- ID of our currently summoned Spirit Link totem, or nil
local isBoss = false                     -- flag to indicate if a boss encounter is in progress
local encounterName = "Unknown"          -- highest classification enemy detected
local archive = { }                      -- table of stored statistics from previous combats
local useAll = true                      -- steering flag to indicate whether to include all encounters (true), or only bosses (false) in grand totals

-- erase current combat data; called when entering or leaving combat
function RShamanStatClear()
	effectiveTotal = 0.0;
    weightedTotal = 0.0;
	masteryTotal = 0.0;
	criticalTotal = 0.0;
	incrementalMastery = 0.0;
	incrementalCritical = 0.0;
	lastCast = { };
	healthTable = { };
	isBoss = false;
	encounterName = "Unknown";
end

-- erase current combat data AND all archived stats
function RShamanStatClearAll()
	archive = { };
	spiritLinkGUID = nil;
	RShamanStatClear();
end

-- initialization
function RShamanStatOnLoad(self)
	local lClass, eClass = UnitClass("player");
	if eClass ~= "SHAMAN" then
		return;
	end
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("UNIT_MAXHEALTH");
	self:RegisterEvent("UNIT_HEALTH_FREQUENT");
	RShamanStatClearAll();
	print(("Resto Shaman Stat Analyzer (r%s) loaded!"):format(revision));
end


function RShamanStatPrintRSM(archive)
    if (#archive == 0) then
        print("RSS: no data available");
        return;
    end

    for i, v in ipairs(archive) do
        if (useAll or v.boss) then
            if v.tot ~= 0 then
                print(("RSS: #%02d - %s: Healing Done: %.1fk, Mastery Effect: %.2f"):format(i, v.label, v.tot/1000, v.weightedTotal/v.tot));
            else
                print(("RSS: #%02d - %s: No Healing Done"):format(i, v.label));
            end
        end
    end
end

function RShamanStatPrint(tot, crtTot, mstTot, crtInc, mstInc)
	print(("RSS: Effective Healing Done: %.1fk"):format(tot/1000));
	if tot > 0 then
		print(("RSS: Throughput from Critical: %.2f%%"):format(100*crtTot/tot));
		print(("RSS: Throughput from Mastery:  %.2f%%"):format(100*mstTot/tot));
		if crtInc > 0 then
			print(("RSS: Critical Rating for +1%% Throughput: %.1f"):format(0.01/(crtInc/tot)));
		end
		if mstInc > 0 then
			print(("RSS: Mastery Rating for +1%% Throughput: %.1f"):format(0.01/(mstInc/tot)));
		end
	end
end

SLASH_RSHAMANSTAT1 = '/rss';
-- Print data from combat that just ended
-- Save it to archive table
-- Zero accumulators for next combat
function RShamanStatReset()
	RShamanStatPrint(effectiveTotal, criticalTotal, masteryTotal, incrementalCritical, incrementalMastery);
	if effectiveTotal > 0 then
		archive[#archive + 1] = { tot = effectiveTotal, crtTotal = criticalTotal, mstTotal = masteryTotal, crtInc = incrementalCritical, mstInc = incrementalMastery,
                                  weightedTotal = weightedTotal,
		                          boss = isBoss, label = encounterName };
	end
	RShamanStatClear();
end

-- Supported commands:
function RShamanStatCmd(msg, ...)
    if (msg == "help" or msg == "") then
        print(("Resto Shaman Stats (r%s) - Available Commands:"):format(revision));
        print(" * /rss sum          Print all data, and grand totals");
        print(" * /rss clr            Print all data, and grand totals, then wipe all data");
        print(" * /rss boss         Only include boss encounter data for grand totals");
        print(" * /rss all            Include all data for grand totals");
        print(" * /rss rsm            Print data of last encounter in RSM mode");

        if (useAll) then
            print("Currently recording all encounters.");
        else
            print("Currently recording boss encounters only.");
        end

        return;
    end

    if (msg == "rsm") then
        RShamanStatPrintRSM(archive);
        return;
    end

    if (msg == "boss") then
		useAll = false;
	elseif (msg == "all") then
		useAll = true;
    end

	if (msg == "sum" or msg == "clr") then
		tot = 0;
		crtTot = 0;
		mstTot = 0;
		crtInc = 0;
		mstInc = 0;
		if (#archive > 0) then
            for i, v in ipairs(archive) do
                if (useAll or v.boss) then
					print(v.label);
					RShamanStatPrint(v.tot, v.crtTotal, v.mstTotal, v.crtInc, v.mstInc);
					tot = tot + v.tot;
					crtTot = crtTot + v.crtTotal;
					mstTot = mstTot + v.mstTotal;
					crtInc = crtInc + v.crtInc;
					mstInc = mstInc + v.mstInc;
				end
			end
		end
		print("--- Summary ---");
		if (#archive > 0 and tot > 0) then
			RShamanStatPrint(tot, crtTot, mstTot, crtInc, mstInc);
		else
			RShamanStatPrint(0,0,0,0,0);
		end
		print("---------------");
	end
	if (msg == "clr") then
		RShamanStatClearAll();
		print("RSS:: All archived data cleared");
	end
end
SlashCmdList["RSHAMANSTAT"] = RShamanStatCmd;

-- returns the name and classification of any hostile target of anyone we heal (boss detection)
function RShamanStatCheckTarget(unit)
	local unitTarget = unit.."-target";
	if ( UnitExists(unitTarget) and UnitAffectingCombat(unitTarget) and UnitCanAttack(unit, unitTarget) ) then
		return UnitName(unitTarget), UnitClassification(unitTarget);
	end
end

-- sets boss flag if worldboss target is detected, and sets encounter name
function RShamanStatCheckBoss(unit)
	local name, classification = RShamanStatCheckTarget(unit);
	if (classification == "worldboss" or (name ~= nil and encounterName == "Unknown")) then 
		encounterName = name;
	end
	if (classification == "worldboss") then
		isBoss = true;
		print("RSS:: Detected boss encounter "..name);
	end
end

function RShamanStatCombatLogEntry(...)
	local playerGUID = UnitGUID("player");
	local timestamp, combatEvent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = select(1, ...);
	
	if sourceGUID == playerGUID and combatEvent and strfind(combatEvent, 'SPELL_SUMMON') then
		local spellId, spellName, spellSchool = select(12, ...);
		if spellId == 98008 then 
			spiritLinkGUID = destGUID;
		end
		return;
	end
	
	if combatEvent and strfind(combatEvent, 'UNIT_DESTROYED') and destGUID == spiritLinkGUID then
		spiritLinkGUID = nil;
		return;
	end
	
	if (sourceGUID == playerGUID or (spiritLinkGUID and sourceGUID == spiritLinkGUID)) and combatEvent and strfind(combatEvent, '_HEAL') and healthTable[destGUID] ~= nil then
		-- Check for boss encounter
		if ( (not isBoss) or (encounterName == "Unknown") ) then
			RShamanStatCheckBoss(destName);
		end
		--if spiritLinkGUID and sourceGUID == spiritLinkGUID then
		--	DEFAULT_CHAT_FRAME:AddMessage("Spirit Link heal detected");
		--end
		-- We healed somebody
		local spellId, spellName, spellSchool = select(12, ...)
		local amount, overhealing, absorbed, critical = select(15, ...)
		local healed = amount - overhealing;                                    -- ignore overhealing
		effectiveTotal = effectiveTotal + healed;                               -- accumulate total effective throughput

		-- Deal with Mastery				
		local myMasteryBonus = 0.03*GetMastery();  								-- this is the % value show on character panel, including Grace of Air
		local unitHealthTable = healthTable[destGUID];							-- find last known values of health and max health
		local hpratio = 1 - (unitHealthTable.currentHealth/unitHealthTable.maxHealth);
		local healBaseAmount = amount/(1+hpratio*myMasteryBonus);               -- the size of the heal with 0% Mastery
		local masteryGain = amount - healBaseAmount;                            -- increase from base amount from Mastery
		local effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);  -- overhealing is not useful
		local incrementalMasteryGain = 0.0;
		if (overhealing == 0) then
			incrementalMasteryGain = 0.00005 * hpratio * healBaseAmount;        -- how much would +1 Mastery Rating increase this heal? (no benefit if already overhealing)
        end
        weightedTotal = weightedTotal + healed * hpratio
		masteryTotal = masteryTotal + effectiveMasteryGain;						-- accumulate gain from current Mastery
		incrementalMastery = incrementalMastery + incrementalMasteryGain;       -- accumulate gain from +1 Mastery
		
		-- Deal with Crit
		local myCriticalChance = GetSpellCritChance(4)/100;                     -- Nature spell critical chance, as shown on character panel, including buffs
		local thisCast = nil;                                                   -- record this as the most recent cast of this spell, for matching with AA
		if (spellId ~= 52752) then                                              -- unless it is AA healing, which can't trigger AA
			thisCast = { };
			thisCast.amount = amount
			thisCast.critChance = myCriticalChance;
			thisCast.critical = false;
		end
		if (critical or (spellId == 52752)) then                                -- special logic for AA or Tidal Waves -> Healing Surge
			if (spellId == 8004) then                                           -- is this a Healing Surge critical?
				if (UnitAura("player", "Tidal Waves") ~= nil) then              -- tested: even if this cast consumed the last TW stack, we still have the buff at this moment
					myCriticalChance = math.min(1.0, myCriticalChance + 0.3);   -- adjust HS critical chance for TW
					thisCast.critChance = myCriticalChance;                     -- record the buffed value of critical chance for use with any AA triggered by this heal
				end
			end
			if (critical) then                                                  -- this cast crit, so figure out how much we gained over the base heal
				thisCast.critical = true;                                       -- record the cast as critical so we can add the triggered AA to the critical category
				local critBaseAmount = amount/2.06;                             -- assume Burning Primal Diamond as metagem
				local critBonusAmount = amount - critBaseAmount;                
				criticalTotal = criticalTotal + math.max(0.0, critBonusAmount - overhealing);  -- record the effective gain from this crit
				-- how much would +1 Critical Rating increase the *expected* (average) effective amount of this cast?
				if (myCriticalChance > 0) then
					incrementalCritical = incrementalCritical + (0.01/600)*math.max(0.0, critBonusAmount - overhealing)/myCriticalChance;
				end
			else -- Ancestral Awakening                                      
				local mindiff = amount;                                         -- Match this AA to the spell that triggered it, based on the 30% relationship
				local sMin = nil;
				local vMin = nil;
				for s, v in pairs(lastCast) do
					if (math.abs(v.amount*0.3 - amount) < mindiff) then
						mindiff = math.abs(v.amount*0.3 - amount);
						vMin = v;
						sMin = s;
					end
				end
				if (vMin ~= nil and vMin.critical) then                        -- if the trigger for this AA was a crit (rather than the set bonus)..
					criticalTotal = criticalTotal + healed;                    -- count it as gain from critical
					if (vMin.critChance > 0) then
						incrementalCritical = incrementalCritical + (0.01/600)*healed/vMin.critChance;  -- how much would +1 Critical Rating increase the expected (average) value of the cast that triggered it?
					end
				end
			end
		end
		if (thisCast ~= nil) then                                              -- store information on this cast for AA matching and accounting
			lastCast[spellName] = thisCast;
		end
	-- done with Mastery and Crit
	end
end
	
function RShamanStatOnEvent(self, event, ...)
	local currentSpec = GetSpecialization()
	local role = currentSpec and select(6, GetSpecializationInfo(currentSpec)) or "NONE"
	if role ~= "HEALER" then
		return;
	end
	if event == "PLAYER_REGEN_DISABLED" then
		-- entering combat; reset counters
		RShamanStatClear();
	end
	if event == "PLAYER_REGEN_ENABLED" then
		-- leaving combat: print status and reset counters
		RShamanStatReset();
	end
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		RShamanStatCombatLogEntry(...);
	end
	if event == "UNIT_HEALTH_FREQUENT" or event == "UNIT_MAXHEALTH" then
		local unitID = select(1,...);
		if unitID == nil then return; end
		if UnitIsPlayer(unitID) or UnitPlayerOrPetInParty(unitID) or UnitPlayerOrPetInRaid(unitID) then
			local unitGUID = UnitGUID(unitID);
			if unitGUID == nil then return; end
			local unitHealthTable = healthTable[unitGUID];
			if unitHealthTable == nil then
				unitHealthTable = { };
				healthTable[unitGUID] = unitHealthTable;
			end
			unitHealthTable.maxHealth = UnitHealthMax(unitID);
			unitHealthTable.currentHealth = UnitHealth(unitID);
			if unitHealthTable.maxHealth < 1 then
				healthTable[unitGUID] = nil;
			end
		end
	end
end